@@ -8,9 +8,11 @@ from collections import abc
 from io import StringIO
 from pathlib import Path
 
-from typing import Callable
+from typing import Awaitable
 from typing import Dict
 from typing import Any
+from typing import Iterator
+from typing import Mapping
 from typing import Optional
 from typing import TextIO
 from typing import Union
@@ -26,9 +28,9 @@ from liquid.exceptions import Error
 if TYPE_CHECKING:  # pragma: no cover
     from liquid import Environment
     from liquid.ast import ParseTree
+    from liquid.loaders import UpToDate
 
 
-# pylint: disable=too-many-instance-attributes
 class BoundTemplate:
     """A liquid template that has been parsed and is bound to a
     :class:`liquid.Environment`.
@@ -58,7 +60,7 @@ class BoundTemplate:
         name: str = "",
         path: Optional[Union[str, Path]] = None,
         globals: Optional[Dict[str, Any]] = None,
-        uptodate: Optional[Callable[[], bool]] = None,
+        uptodate: UpToDate = None,
     ):
         self.env = env
         self.tree = parse_tree
@@ -67,28 +69,36 @@ class BoundTemplate:
         self.path = path
         self.uptodate = uptodate
 
-    def render(self, *args: ..., **kwargs: ...) -> str:
+    def render(self, *args: Any, **kwargs: Any) -> str:
         """Render the template with `args` and `kwargs` included in the render context.
 
         Accepts the same arguments as the :class:`dict` constructor.
         """
         _vars: Dict[str, object] = dict(*args, **kwargs)
-
         context = Context(self.env, ChainMap(_vars, self.globals))
 
         buf = StringIO()
         self.render_with_context(context, buf)
         return buf.getvalue()
 
+    async def render_async(self, *args: Any, **kwargs: Any) -> str:
+        """An async version of :meth:`liquid.template.BoundTemplate.render`."""
+        _vars: Dict[str, object] = dict(*args, **kwargs)
+        context = Context(self.env, ChainMap(_vars, self.globals))
+
+        buf = StringIO()
+        await self.render_with_context_async(context, buf)
+        return buf.getvalue()
+
     def render_with_context(
         self,
         context: Context,
         buffer: TextIO,
-        *args: ...,
+        *args: Any,
         partial: bool = False,
         block_scope: bool = False,
-        **kwargs: ...,
-    ):
+        **kwargs: Any,
+    ) -> None:
         """Render the template using an existing context and output buffer.
 
         ``args`` and ``kwargs`` are passed to the :class:`dict` constructor. The
@@ -118,14 +128,49 @@ class BoundTemplate:
                     if not partial or block_scope:
                         self.env.error(
                             LiquidSyntaxError(
-                                f"unexpected '{err}'", linenum=node.tok.linenum
+                                f"unexpected '{err}'", linenum=node.token().linenum
                             )
                         )
                     else:
                         raise
                 except Error as err:
                     # Raise or warn according to the current mode.
-                    self.env.error(err, linenum=node.tok.linenum)
+                    self.env.error(err, linenum=node.token().linenum)
+
+    async def render_with_context_async(
+        self,
+        context: Context,
+        buffer: TextIO,
+        *args: Any,
+        partial: bool = False,
+        block_scope: bool = False,
+        **kwargs: Any,
+    ) -> None:
+        """An async version of :meth:`liquid.template.BoundTemplate.render_with_context`."""
+        # "template" could get overridden from args/kwargs, "partial" will not.
+        namespace = self._make_globals(partial, args, kwargs)
+
+        with context.extend(namespace=namespace):
+            for node in self.tree.statements:
+                try:
+                    await node.render_async(context, buffer)
+                except LiquidInterrupt as err:
+                    # If this is an "included" template, there could be a for loop
+                    # in a parent template. A for loop that could be interrupted
+                    # from an included template.
+                    #
+                    # Convert the interrupt to a syntax error if there is no parent.
+                    if not partial or block_scope:
+                        self.env.error(
+                            LiquidSyntaxError(
+                                f"unexpected '{err}'", linenum=node.token().linenum
+                            )
+                        )
+                    else:
+                        raise
+                except Error as err:
+                    # Raise or warn according to the current mode.
+                    self.env.error(err, linenum=node.token().linenum)
 
     @property
     def is_up_to_date(self) -> bool:
@@ -133,15 +178,37 @@ class BoundTemplate:
         True otherwise."""
         if not self.uptodate:
             return True
-        return self.uptodate()
 
-    def _make_globals(self, partial, args, kwargs) -> abc.Mapping:
+        uptodate = self.uptodate()
+        if not isinstance(uptodate, bool):
+            raise Error(
+                f"expected a boolean from uptodate, found {type(uptodate).__name__}"
+            )
+        return uptodate
+
+    async def is_up_to_date_async(self) -> bool:
+        """An async version of the ``is_up_to_date`` property.
+
+        If ``template.uptodate`` is a coroutine, it wil be awaited. Otherwise it will be
+        called just like ``is_up_to_date``
+        """
+        if not self.uptodate:
+            return True
+
+        uptodate = self.uptodate()
+        if isinstance(uptodate, Awaitable):
+            return await uptodate
+        return uptodate
+
+    def _make_globals(
+        self, partial: bool, args: Any, kwargs: Any
+    ) -> abc.Mapping[str, object]:
         return {
             **dict(*args, **kwargs),
             "partial": partial,
         }
 
-    def __repr__(self):
+    def __repr__(self) -> str:
         return (
             f"Template(name='{self.name}', "
             f"path='{self.path}', uptodate={self.is_up_to_date})"
@@ -149,15 +216,17 @@ class BoundTemplate:
 
 
 class AwareBoundTemplate(BoundTemplate):
-    def __init__(self, *args, **kwargs):
+    def __init__(self, *args: Any, **kwargs: Any) -> None:
         super().__init__(*args, **kwargs)
         self.drop = TemplateDrop(self.name, self.path)
 
-    def _make_globals(self, partial, args, kwargs) -> abc.Mapping:
+    def _make_globals(
+        self, partial: bool, args: Any, kwargs: Any
+    ) -> abc.Mapping[str, object]:
         return {"template": self.drop, **super()._make_globals(partial, args, kwargs)}
 
 
-class TemplateDrop(abc.Mapping):
+class TemplateDrop(Mapping[str, Optional[str]]):
     """Template meta data mapping."""
 
     def __init__(self, name: str, path: Optional[Union[str, Path]]):
@@ -168,10 +237,10 @@ class TemplateDrop(abc.Mapping):
             self.path = Path(self.name)
 
         self.stem = self.path.stem
+        self.suffix: Optional[str] = None
+
         if "." in self.stem:
             self.suffix = self.stem.split(".")[-1]
-        else:
-            self.suffix = None
 
         self._items = {
             "directory": self.path.parent.name,
@@ -179,23 +248,23 @@ class TemplateDrop(abc.Mapping):
             "suffix": self.suffix,
         }
 
-    def __str__(self):
+    def __str__(self) -> str:
         return self.stem
 
-    def __repr__(self):
+    def __repr__(self) -> str:
         return (
             f"TemplateDrop(directory='{self['directory']}', "
             f"name='{self['name']}', suffix='{self['suffix']}')"
         )  # pragma: no cover
 
-    def __contains__(self, item):
+    def __contains__(self, item: object) -> bool:
         return item in self._items
 
-    def __getitem__(self, key):
-        return self._items[key]
+    def __getitem__(self, key: object) -> Optional[str]:
+        return self._items[str(key)]
 
-    def __len__(self):
+    def __len__(self) -> int:
         return len(self._items)
 
-    def __iter__(self):
+    def __iter__(self) -> Iterator[str]:
         return iter(self._items)
