@@ -202,6 +202,12 @@ class RenderTestCases(unittest.TestCase):
                     expect="garden",
                     globals={"product": {"tags": ["sports", "garden"]}},
                 ),
+                Case(
+                    description="array item by negative index",
+                    template=r"{{ product.tags[-2] }}",
+                    expect="sports",
+                    globals={"product": {"tags": ["sports", "garden"]}},
+                ),
                 Case(
                     description="array item by index from identifier",
                     template="{% assign i = 1 %}{{ product.tags[i] }}",
@@ -515,6 +521,16 @@ class RenderTestCases(unittest.TestCase):
                 expect="username is blank",
                 globals={"username": ""},
             ),
+            Case(
+                description="conditional alternative with default",
+                template=(
+                    r"{% if false %}foo"
+                    r"{% elsif false %}bar"
+                    r"{% else %}hello"
+                    r"{% endif %}"
+                ),
+                expect="hello",
+            ),
         ]
 
         self._test(test_cases)
@@ -562,6 +578,26 @@ class RenderTestCases(unittest.TestCase):
                 template=r"{% unless false %}  {% endunless %}",
                 expect="",
             ),
+            Case(
+                description="alternative",
+                template=r"{% unless true %}foo{% else %}bar{% endunless %}",
+                expect="bar",
+            ),
+            Case(
+                description="conditional alternative",
+                template=r"{% unless true %}foo{% elsif true %}bar{% endunless %}",
+                expect="bar",
+            ),
+            Case(
+                description="conditional alternative with default",
+                template=(
+                    r"{% unless true %}foo"
+                    r"{% elsif false %}bar"
+                    r"{% else %}hello"
+                    r"{% endunless %}"
+                ),
+                expect="hello",
+            ),
         ]
 
         self._test(test_cases)
@@ -750,12 +786,12 @@ class RenderTestCases(unittest.TestCase):
             Case(
                 description="named counter",
                 template=r"{% decrement foo %}{{ foo }} {% decrement foo %}{{ foo }}",
-                expect="-1 -2",
+                expect="-1-1 -2-2",
             ),
             Case(
                 description="increment and decrement named counter",
                 template=r"{% decrement foo %} {% decrement foo %} {% increment foo %}",
-                expect="-1 -2 -1",
+                expect="-1 -2 -2",
             ),
         ]
 
@@ -770,6 +806,11 @@ class RenderTestCases(unittest.TestCase):
                 template=r"{% increment foo %} {% increment foo %} {% increment foo %}",
                 expect="0 1 2",
             ),
+            Case(
+                description="incrementing counter renders before incrementing",
+                template=r"{% increment foo %} {{ foo }}",
+                expect="0 1",
+            ),
             Case(
                 description="multiple named counters",
                 template=(
@@ -791,6 +832,17 @@ class RenderTestCases(unittest.TestCase):
                 ),
                 expect="5 0 1 5",
             ),
+            Case(
+                description="named counters are in scope for subsequent expressions",
+                template=(
+                    r"{% increment foo %} "
+                    r"{% increment foo %} "
+                    r"{% if foo > 0 %}"
+                    r"{{ foo }}"
+                    r"{% endif %}"
+                ),
+                expect="0 1 2",
+            ),
         ]
 
         self._test(test_cases)
@@ -1102,6 +1154,60 @@ class RenderTestCases(unittest.TestCase):
                 expect="1 2 3 4 5 6 ",
                 globals={"array": [1, 2, 3, 4, 5, 6]},
             ),
+            Case(
+                description="parentloop is normally undefined",
+                template=(
+                    r"{% for i in (1..2)%}{{ forloop.parentloop.index }}{% endfor %}"
+                ),
+                expect="",
+                globals={},
+            ),
+            Case(
+                description="access parentloop",
+                template=(
+                    r"{% for i in (1..2)%}"
+                    r"{% for j in (1..2) %}"
+                    r"{{ i }} {{j}} {{ forloop.parentloop.index }} {{ forloop.index }} "
+                    r"{% endfor %}"
+                    r"{% endfor %}"
+                ),
+                expect="1 1 1 1 1 2 1 2 2 1 2 1 2 2 2 2 ",
+                globals={},
+            ),
+            Case(
+                description="parentloop goes out of scope",
+                template=(
+                    r"{% for i in (1..2)%}"
+                    r"{% for j in (1..2) %}"
+                    r"{{ i }} {{ j }} "
+                    r"{% endfor %}"
+                    r"{{ forloop.parentloop.index }}"
+                    r"{% endfor %}"
+                ),
+                expect="1 1 1 2 2 1 2 2 ",
+                globals={},
+            ),
+            Case(
+                description="parent's parentloop",
+                template=(
+                    r"{% for i in (1..2) %}"
+                    r"{% for j in (1..2) %}"
+                    r"{% for k in (1..2) %}"
+                    r"i={{ forloop.parentloop.parentloop.index }} "
+                    r"j={{ forloop.parentloop.index }} "
+                    r"k={{ forloop.index }} "
+                    r"{% endfor %}"
+                    r"{% endfor %}"
+                    r"{% endfor %}"
+                ),
+                expect=(
+                    "i=1 j=1 k=1 i=1 j=1 k=2 "
+                    "i=1 j=2 k=1 i=1 j=2 k=2 "
+                    "i=2 j=1 k=1 i=2 j=1 k=2 "
+                    "i=2 j=2 k=1 i=2 j=2 k=2 "
+                ),
+                globals={},
+            ),
         ]
 
         self._test(test_cases)
@@ -1679,6 +1785,40 @@ class RenderTestCases(unittest.TestCase):
                 globals={"loop": MockIterableDrop()},
                 partials={"loop-with": r"{{ value }}"},
             ),
+            Case(
+                description="render loops don't add parentloop",
+                template=r"{% render 'product' for collection.products %}",
+                expect="bike-0 0 1 2 car-1 0 1 2 ",
+                globals={
+                    "collection": {"products": [{"title": "bike"}, {"title": "car"}]}
+                },
+                partials={
+                    "product": (
+                        r"{{ product.title }}-{{ forloop.index0 }} "
+                        r"{% for x in (1..3) %}"
+                        r"{{ forloop.index0 }}{{ forloop.parentloop.index0 }} "
+                        r"{% endfor %}"
+                    ),
+                },
+            ),
+            Case(
+                description="render loops can't access parentloop",
+                template=(
+                    r"{% for x in (1..3) %}"
+                    r"{% render 'product' for collection.products %}"
+                    r"{% endfor %}"
+                ),
+                expect="bike-0 car-1 bike-0 car-1 bike-0 car-1 ",
+                globals={
+                    "collection": {"products": [{"title": "bike"}, {"title": "car"}]}
+                },
+                partials={
+                    "product": (
+                        r"{{ product.title }}-{{ forloop.index0 }} "
+                        r"{{ forloop.parentloop.index0 }}"
+                    ),
+                },
+            ),
         ]
 
         self._test(test_cases)
